import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.net.Socket;
import java.security.DigestException;
import java.util.concurrent.LinkedBlockingQueue;

import javax.swing.JFrame;

/*
 *  Thread to perform auto sync functions
 */
class AutoSyncThread {
	int time;
	boolean dismissed = false;
	LinkedBlockingQueue<Msg> inbox;
	MainFrame mf;

	AutoSyncThread(int time, LinkedBlockingQueue<Msg> inbox, MainFrame mf) {
		this.time = time;
		this.inbox = inbox;
		this.mf = mf;
		start();
	}

	void start() {
		new Thread(new Runnable() {

			@Override
			public void run() {
				int t = time;
				while (t > 0 && !dismissed) {
					mf.updateOutput(null, "Auto syncing up in " + t + " seconds.");
					try {
						Thread.sleep(1000);
					} catch (InterruptedException e) {
						e.printStackTrace();
					}
					t--;
				}
				// update the gui
				mf.updateOutput(null, "");
				if (!dismissed) {
					Msg msg = new Msg(Msg.SYNC_UP_ALARM, null);
					inbox.add(msg);
				}
			}
		}).start();
	}
}

/**
 * 
 * This class implements all the client side functionality.
 *
 */
public class Client {
	String configFileName;
	String serverAddr;
	int serverPort;

	Connection connection = null;
	boolean syncGoingOn = false;
	boolean loggedIn = false;

	String parentDir;

	static String mainFolderName = "FSync";
	static String tempFolderName = ".temp";

	File sigFile = null, deltaFile = null;
	DataOutputStream sigFileOut = null, deltaFileOut = null;

	JFrame gui = null;

	int bytesTransferred = 0;

	LinkedBlockingQueue<Msg> inbox = new LinkedBlockingQueue<Msg>();

	AutoSyncThread autoSyncThread = null;

	double round(double a) {
		return Math.round(a * 1000.0) / 1000.0;
	}

	/**
	 * Reads the configuration file and sets the server address, port and working
	 * directory.
	 */
	public void readConfig() throws IOException {
		BufferedReader in = new BufferedReader(new FileReader(new File(configFileName)));
		while (true) {
			String line = in.readLine();
			if (line == null)
				break;
			if (line.length() < 1 || line.startsWith("#"))
				continue;
			String tokens[] = line.split(":");
			if (tokens[0].equals("SERVER_ADDR")) {
				serverAddr = tokens[1];
			} else if (tokens[0].equals("SERVER_PORT")) {
				serverPort = Integer.parseInt(tokens[1]);
			} else if (tokens[0].equals("OPERATION_DIR")) {
				parentDir = tokens[1];
			}
		}
		in.close();
	}

	/**
	 * Sends login request.
	 * 
	 * @param msg : this contains the username and password
	 * @throws IOException
	 */
	public void tryToLogin(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		String uname = din.readUTF();
		String pass = din.readUTF();

		Msg m = new Msg(Msg.OUTGOING_MSG, null);
		DataOutputStream dout = m.getDataOutputStream();
		dout.writeInt(Msg.LOGIN);
		dout.writeUTF(uname);
		dout.writeUTF(pass);
		m.prepare();
		connection.send(m);
	}

	/**
	 * Processes login request result from the server.
	 * 
	 * @param msg : server's reply
	 * @throws IOException
	 */
	public void handleLoginReply(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		int res = din.readInt();
		if (res == Msg.LOGIN_ERROR) {
			// login failed
			String error = din.readUTF();
			System.out.println(error);
			LoginFrame lf = (LoginFrame) gui;
			lf.updateOutput(error);
		} else {
			// logged in
			String uname = din.readUTF();
			System.out.println("logged in as : " + uname);
			loggedIn = true;
			int x = gui.getX(), y = gui.getY();
			gui.setVisible(false);
			gui = new MainFrame(inbox, "Idle", "", "Logged in as : \n" + uname);
			((MainFrame) gui).updateAutoSyncStatus(false, 0);
			gui.setLocation(x, y);
			gui.setVisible(true);
		}
	}

	/**
	 * Send signup request.
	 * 
	 * @param msg : this contains username and password
	 * @throws IOException
	 */
	public void tryToSignup(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		String uname = din.readUTF();
		String pass = din.readUTF();

		Msg m = new Msg(Msg.OUTGOING_MSG, null);
		DataOutputStream dout = m.getDataOutputStream();
		dout.writeInt(Msg.SIGNUP);
		dout.writeUTF(uname);
		dout.writeUTF(pass);
		m.prepare();
		connection.send(m);
	}

	/**
	 * Processes signup request result from the server.
	 * 
	 * @param msg : server's reply
	 * @throws IOException
	 */
	public void handleSignupReply(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		int res = din.readInt();
		if (res == Msg.SIGNUP_ERROR) {
			String error = din.readUTF();
			System.out.println(error);
			SignupFrame sf = (SignupFrame) gui;
			sf.updateOutput(error);
		} else {
			System.out.println("signed up");
			int x = gui.getX(), y = gui.getY();
			gui.setVisible(false);
			gui = new LoginFrame(inbox, "Signup complete. Login using your username and password.");
			gui.setLocation(x, y);
			gui.setVisible(true);
		}
	}

	/**
	 * Send the sync up request to the server.
	 * 
	 * @throws IOException
	 */
	public void initiateSyncup() throws IOException {
		if (!loggedIn || syncGoingOn) {
			if (!loggedIn)
				System.out.println("not logged in");
			else if (syncGoingOn)
				System.out.println("sync going on");
			return;
		}
		if (autoSyncThread != null)
			autoSyncThread.dismissed = true;
		System.out.println("sending sync up req");
		MainFrame mf = (MainFrame) gui;
		mf.updateOutput("Trying to Sync Up...", "");
		syncGoingOn = true;
		Msg m = new Msg(Msg.OUTGOING_MSG, null);
		DataOutputStream dout = m.getDataOutputStream();
		dout.writeInt(Msg.INIT_SYNC_UP_REQ_FROM_CLIENT);
		m.prepare();
		connection.send(m);
	}

	/**
	 * Send the sync down request to the server.
	 * 
	 * @throws IOException
	 */
	public void initiateSyncdown() throws IOException {
		if (!loggedIn || syncGoingOn) {
			if (!loggedIn)
				System.out.println("not logged in");
			else if (syncGoingOn)
				System.out.println("sync going on");
			return;
		}
		System.out.println("sending sync down req");
		MainFrame mf = (MainFrame) gui;
		mf.updateOutput("Trying to Sync Down...", "");
		syncGoingOn = true;
		Msg m = new Msg(Msg.OUTGOING_MSG, null);
		DataOutputStream dout = m.getDataOutputStream();
		dout.writeInt(Msg.INIT_SYNC_DOWN_REQ_FROM_CLIENT);
		m.prepare();
		connection.send(m);
	}

	/**
	 * This is called when the sync up request is denied.
	 * 
	 * @param msg : server's reply
	 * @throws IOException
	 */
	public void handleSyncUpReqDenied(Msg msg) throws IOException {
		syncGoingOn = false;
		DataInputStream din = msg.getDataInputStream();
		String error = din.readUTF();
		System.out.println(error);
		MainFrame mf = (MainFrame) gui;
		mf.updateOutput(error, "");
		if (autoSyncThread != null) {
			autoSyncThread = new AutoSyncThread(autoSyncThread.time, inbox, mf);
		}
	}

	/**
	 * This is called when the sync down request is denied.
	 * 
	 * @param msg : server's reply
	 * @throws IOException
	 */
	public void handleSyncDownReqDenied(Msg msg) throws IOException {
		syncGoingOn = false;
		DataInputStream din = msg.getDataInputStream();
		String error = din.readUTF();
		System.out.println(error);
		MainFrame mf = (MainFrame) gui;
		mf.updateOutput(error, "");
	}

	/**
	 * Starts the sync down process. This is called when the client gets an approval
	 * for sync down request or an instruction to perform sync down from server.
	 * 
	 * @throws IOException
	 * @throws DigestException
	 */
	public void performSyncDown() throws IOException, DigestException {
		bytesTransferred = 0;
		System.out.println("performing sync down");
		MainFrame mf = (MainFrame) gui;
		mf.updateOutput("Syncing Down...", "");
		File aio = new File(parentDir + tempFolderName + "/aio");
		FileOps.createAIOFile(new File(parentDir + mainFolderName), aio);
		Sync sync = new Sync();
		sigFile = new File(parentDir + tempFolderName + "/sigfile");
		sync.generateSigFile(aio, sigFile);
		connection.sendFile(sigFile, Msg.SIGFILE);
		bytesTransferred += sigFile.length();
		sigFile.delete();
		sigFile = null;
	}

	/**
	 * This is called when client receives a part of the signature file.
	 * 
	 * @param msg : contains a part of the signature file
	 * @throws IOException
	 * @throws InvalidSignatureFile
	 */
	public void processSigFile(Msg msg) throws IOException, InvalidSignatureFile {
		if (sigFile == null) {
			// first part is received
			bytesTransferred = 0;
			MainFrame mf = (MainFrame) gui;
			System.out.println("Syncing Up...");
			mf.updateOutput("Syncing Up...", "");
			sigFile = new File(parentDir + tempFolderName + "/sigfile");
			sigFileOut = new DataOutputStream(new FileOutputStream(sigFile));
		}
		DataInputStream din = msg.getDataInputStream();
		int isLastPart = din.readInt();
		int partSize = din.readInt();

		byte b[] = new byte[1024];
		while (partSize > 0) {
			int read = din.read(b);
			sigFileOut.write(b, 0, read);
			partSize -= read;
		}

		if (isLastPart == 1) {
			// sig file completely received
			System.out.println("sig file received\nSending delta file");
			sigFileOut.close();
			File aio = new File(parentDir + tempFolderName + "/aio");
			FileOps.createAIOFile(new File(parentDir + mainFolderName), aio);
			Sync sync = new Sync();
			deltaFile = new File(parentDir + tempFolderName + "/deltafile");
			sync.generateDeltaFile(aio, sigFile, deltaFile);
			connection.sendFile(deltaFile, Msg.DELTAFILE);
			aio.delete();
			bytesTransferred += sigFile.length();
			bytesTransferred += deltaFile.length();
			sigFile.delete();
			sigFile = null;
			deltaFile.delete();
			deltaFile = null;
			syncGoingOn = false;
			System.out.println("Sync Up complete.\n" + round(bytesTransferred / 1024.0) + " KB transferred.");
			MainFrame mf = (MainFrame) gui;
			mf.updateOutput("Sync Up complete.\n" + round(bytesTransferred / 1024.0) + " KB transferred.", "");
			if (autoSyncThread != null) {
				autoSyncThread = new AutoSyncThread(autoSyncThread.time, inbox, autoSyncThread.mf);
			}
		}
	}

	/**
	 * This is called when client receives a part of the delta file.
	 * 
	 * @param msg : contains a part of the delta file
	 * @throws IOException
	 * @throws InvalidSignatureFile
	 */
	public void processDeltaFile(Msg msg) throws IOException, InvalidSignatureFile {
		if (deltaFile == null) {
			// first part is received
			deltaFile = new File(parentDir + tempFolderName + "/deltafile");
			deltaFileOut = new DataOutputStream(new FileOutputStream(deltaFile));
		}
		DataInputStream din = msg.getDataInputStream();
		int isLastPart = din.readInt();
		int partSize = din.readInt();

		byte b[] = new byte[1024];
		while (partSize > 0) {
			int read = din.read(b);
			deltaFileOut.write(b, 0, read);
			partSize -= read;
		}

		if (isLastPart == 1) {
			// delta file completely received
			System.out.println("delta file received");
			deltaFileOut.close();
			File aio = new File(parentDir + tempFolderName + "/aio");
			File aio2 = new File(parentDir + tempFolderName + "/aio2");
			Sync sync = new Sync();
			sync.applyDelta(aio, deltaFile, aio2);
			FileOps.deleteDir(new File(parentDir + mainFolderName));
			FileOps.createFolderFromAIOFile(aio2, parentDir);
			aio.delete();
			aio2.delete();
			bytesTransferred += deltaFile.length();
			deltaFile.delete();
			deltaFile = null;
			syncGoingOn = false;
			System.out.println("Sync Down complete.\n" + round(bytesTransferred / 1024.0) + " KB transferred.");
			MainFrame mf = (MainFrame) gui;
			mf.updateOutput("Sync Down complete.\n" + round(bytesTransferred / 1024.0) + " KB transferred.", "");
		}
	}

	/**
	 * Reads the message name of the incoming message and calls the appropriate
	 * handler.
	 * 
	 * @param msg : incoming message
	 * @throws IOException
	 * @throws DigestException
	 * @throws InvalidSignatureFile
	 */
	public void handleIncomingMsg(Msg msg) throws IOException, DigestException, InvalidSignatureFile {
		DataInputStream din = msg.getDataInputStream();
		int msgName = din.readInt();
		if (msgName == Msg.LOGIN_RESULT) {
			handleLoginReply(msg);
		} else if (msgName == Msg.SIGNUP_RESULT) {
			handleSignupReply(msg);
		} else if (msgName == Msg.SYNC_DOWN_REQ_DENIED) {
			handleSyncDownReqDenied(msg);
		} else if (msgName == Msg.SYNC_UP_REQ_DENIED) {
			handleSyncUpReqDenied(msg);
		} else if (msgName == Msg.SYNC_DOWN_REQ_GO_AHEAD) {
			performSyncDown();
		} else if (msgName == Msg.SIGFILE) {
			processSigFile(msg);
		} else if (msgName == Msg.DELTAFILE) {
			processDeltaFile(msg);
		} else if (msgName == Msg.INIT_SYNC_DOWN_INSTRUCTION_FROM_SERVER) {
			performSyncDown();
		}
	}

//	void startInputThread(LinkedBlockingQueue<Msg> inbox) {
//		new Thread(new Runnable() {
//
//			@Override
//			public void run() {
//				BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
//				while (true) {
//					String line;
//					try {
//						line = br.readLine();
//						if (line == null)
//							continue;
//						if (line.equals("l")) {
//							String uname = br.readLine();
//							String pass = br.readLine();
//							Msg m = new Msg(Msg.LOGIN_ALARM, null);
//							DataOutputStream dout = m.getDataOutputStream();
//							dout.writeUTF(uname);
//							dout.writeUTF(pass);
//							m.prepare();
//							inbox.add(m);
//						} else if (line.equals("s")) {
//							String uname = br.readLine();
//							String pass = br.readLine();
//							Msg m = new Msg(Msg.SIGNUP_ALARM, null);
//							DataOutputStream dout = m.getDataOutputStream();
//							dout.writeUTF(uname);
//							dout.writeUTF(pass);
//							m.prepare();
//							inbox.add(m);
//						} else if (line.equals("up")) {
//							Msg m = new Msg(Msg.SYNC_UP_ALARM, null);
//							inbox.add(m);
//						} else if (line.equals("down")) {
//							Msg m = new Msg(Msg.SYNC_DOWN_ALARM, null);
//							inbox.add(m);
//						}
//					} catch (IOException e) {
//						e.printStackTrace();
//					}
//				}
//			}
//		}).start();
//	}

	/**
	 * Reads config file, connects to the server and creates the client side GUI.
	 * 
	 * @throws IOException
	 */
	public void start() throws IOException {
		readConfig();
		File f = new File(parentDir + mainFolderName);
		f.mkdir();
		f = new File(parentDir + tempFolderName);
		f.mkdir();

		Socket socket = null;
		try {
			socket = new Socket(serverAddr, serverPort);
		} catch (IOException e) {
			System.out.println("Unable to connect to server! Exiting.");
			System.exit(0);
		}

		connection = new Connection(socket, inbox);
		connection.startRecSendThreads();

		gui = new LoginFrame(inbox, "Login with your username and password.");
		gui.setLocation(200, 200);
		gui.setVisible(true);

		while (true) {
			try {
				Msg msg = inbox.take();
				if (msg.type == Msg.CONNECTION_ENDED) {
					System.out.println("Connection from server ended! Exiting.");
					System.exit(0);
				} else if (msg.type == Msg.LOGIN_ALARM) {
					tryToLogin(msg);
				} else if (msg.type == Msg.SIGNUP_ALARM) {
					tryToSignup(msg);
				} else if (msg.type == Msg.GOTO_LOGIN) {
					int x = gui.getX(), y = gui.getY();
					gui.setVisible(false);
					gui = new LoginFrame(inbox, "Login with your username and password.");
					gui.setLocation(x, y);
					gui.setVisible(true);
				} else if (msg.type == Msg.GOTO_SIGNUP) {
					int x = gui.getX(), y = gui.getY();
					gui.setVisible(false);
					gui = new SignupFrame(inbox, "Create a new account.");
					gui.setLocation(x, y);
					gui.setVisible(true);
				} else if (msg.type == Msg.TURN_OFF_AUTO_SYNC) {
					autoSyncThread.dismissed = true;
					autoSyncThread = null;
					MainFrame mf = (MainFrame) gui;
					mf.updateAutoSyncStatus(false, 0);
					mf.updateOutput(null, "");
				} else if (msg.type == Msg.TURN_ON_AUTO_SYNC) {
					MainFrame mf = (MainFrame) gui;
					mf.updateAutoSyncStatus(true, (Integer) msg.userData);
					autoSyncThread = new AutoSyncThread((Integer) msg.userData, inbox, mf);
				} else if (msg.type == Msg.SYNC_UP_ALARM) {
					initiateSyncup();
				} else if (msg.type == Msg.SYNC_DOWN_ALARM) {
					initiateSyncdown();
				} else if (msg.type == Msg.INCOMING_MSG) {
					try {
						handleIncomingMsg(msg);
					} catch (DigestException e) {
						e.printStackTrace();
					} catch (InvalidSignatureFile e) {
						e.printStackTrace();
					}
				}

			} catch (InterruptedException e) {
				e.printStackTrace();
			}

		}
	}

	public static void main(String[] args) throws InterruptedException {
		if (args.length < 1) {
			System.out.println("No config file provided in arguments!");
			return;
		}
		Client client = new Client();
		client.configFileName = args[0];
		try {
			client.start();
		} catch (IOException e) {
			System.out.println("Connection from server broken! Exiting.");
			e.printStackTrace();
		}
	}
}
