import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

/**
 * 
 * Represents a message sent either over the network or internally. Defines all
 * the message types and message names.
 *
 */
public class Msg {
	final static int CONNECTION_ENDED = 1;
	final static int INCOMING_MSG = 2;
	final static int OUTGOING_MSG = 3;
	final static int ALARM = 4;
	final static int SYNC_THREAD_ENDED = 5;
	final static int END_SYNC_ALARM = 6;

	final static int SIGNUP = 100;
	final static int LOGIN = 101;
	final static int SIGNUP_RESULT = 200;
	final static int LOGIN_RESULT = 201;

	final static int TURN_ON_AUTO_SYNC = 301;
	final static int TURN_OFF_AUTO_SYNC = 302;

	final static int LOGIN_ERROR = 1000;
	final static int LOGIN_SUCCSESS = 1001;
	final static int SIGNUP_ERROR = 1002;
	final static int SIGNUP_SUCCSESS = 1003;

	final static int SIGNUP_ALARM = 2001;
	final static int LOGIN_ALARM = 2002;
	final static int SYNC_UP_ALARM = 2003;
	final static int SYNC_DOWN_ALARM = 2004;
	final static int GOTO_LOGIN = 2005;
	final static int GOTO_SIGNUP = 2006;

	final static int INIT_SYNC_UP_REQ_FROM_CLIENT = 3001;
	final static int INIT_SYNC_DOWN_REQ_FROM_CLIENT = 3002;
	final static int INIT_SYNC_DOWN_INSTRUCTION_FROM_SERVER = 3003;

	final static int SYNC_UP_REQ_DENIED = 4001;
	final static int SYNC_DOWN_REQ_DENIED = 4002;
	final static int SYNC_DOWN_REQ_GO_AHEAD = 4003;

	final static int SIGFILE = 5001;
	final static int DELTAFILE = 5002;

	int type;
	byte msg[];
	Connection connection;
	Object userData; // can be used to store arbitrary data

	private DataInputStream din = null;
	private DataOutputStream dout = null;
	private ByteArrayOutputStream bout = null;

	/**
	 * @param type       : Message type
	 * @param msg        : Message content
	 * @param connection : network connection associated with the message
	 */
	public Msg(int type, byte msg[], Connection connection) {
		this.type = type;
		this.msg = msg;
		this.connection = connection;
	}

	/**
	 * @param type     : Message type
	 * @param userData : Any arbitrary data to be associated with message
	 */
	public Msg(int type, Object userData) {
		this.type = type;
		this.userData = userData;
	}

	/**
	 * Resets the input stream of the message
	 */
	public void resetDataInputStream() {
		din = null;
	}

	/**
	 * Get the input stream to read the message content.
	 * 
	 * @return Input stream of content of the message
	 */
	public DataInputStream getDataInputStream() {
		if (din == null) {
			din = new DataInputStream(new ByteArrayInputStream(msg));
		}
		return din;
	}

	/**
	 * Get the output stream to write the message content.
	 * 
	 * @return Output stream for content of the message
	 */
	public DataOutputStream getDataOutputStream() {
		if (dout == null) {
			bout = new ByteArrayOutputStream();
			dout = new DataOutputStream(bout);
		}
		return dout;
	}

	/**
	 * Call this method when done writing to the output stream of the message to
	 * prepare the message to be sent out.
	 * 
	 * @throws IOException
	 */
	public void prepare() throws IOException {
		dout.flush();
		bout.flush();
		msg = bout.toByteArray();
	}

}
