import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Random;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

/**
 * 
 * Represents a network connection.
 *
 */

public class Connection {
	Socket socket;
	DataInputStream dataIn;
	DataOutputStream dataOut;

	// all incoming msgs go to inQueue
	// all msgs that are placed in outQueue are sent through this connection's
	// socket
	LinkedBlockingQueue<Msg> inQueue, outQueue;
	Thread recThread, sendThread;

	Object userData; // can be used to store arbitrary data

	boolean isActive = false, keepRunning = false;

	String userName = null;

	File sigFile = null, deltaFile = null;
	DataOutputStream sigFileOut = null, deltaFileOut = null;

	/**
	 * 
	 * @param socket  : network socket for this connection
	 * @param inQueue : main thread's inbox, all incoming messages will be put in
	 *                this queue
	 * @throws IOException
	 */
	public Connection(Socket socket, LinkedBlockingQueue<Msg> inQueue) throws IOException {
		this.socket = socket;
		dataIn = new DataInputStream(socket.getInputStream());
		dataOut = new DataOutputStream(socket.getOutputStream());
		this.inQueue = inQueue;
		outQueue = new LinkedBlockingQueue<>();
	}

	/**
	 * Adds the given message to the send queue.
	 * 
	 * @param msg : message to be sent
	 */
	public void send(Msg msg) {
		outQueue.add(msg);
	}

	/**
	 * Send the whole file through this connection.
	 * 
	 * @param f        : file to be sent
	 * @param fileType : either SIGFILE or DELTAFILE
	 * @throws IOException
	 */
	public void sendFile(File f, int fileType) throws IOException {
		DataInputStream din = new DataInputStream(new FileInputStream(f));
		long size = f.length();
		byte b[] = new byte[1024];
		while (size > 0) {
			int read = din.read(b);
			size -= read;

			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(fileType);
			if (size == 0)
				dout.writeInt(1);
			else
				dout.writeInt(0);
			dout.writeInt(read);
			dout.write(b, 0, read);
			m.prepare();
			send(m);
		}
		din.close();
	}

	/**
	 * Creates send and receive threads. Send thread sends all the messages that it
	 * receives from other parts of the program through this connection. Receive
	 * thread receives all the messages from this connection and passes them on to
	 * main thread.
	 */
	public void startRecSendThreads() {
		Connection connection = this;
		keepRunning = true;
		// start the receiving thread
		recThread = new Thread(new Runnable() {
			public void run() {
				try {
					while (connection.keepRunning) {
						// Msg format [msgLen(32 bits)][msg(msgLen bytes)]
						int len = dataIn.readInt();
						byte msg[] = new byte[len];
						int off = 0;
						while (len > 0) {
							int read = dataIn.read(msg, off, len);
							if (read < 1)
								throw new Exception();
							off += read;
							len -= read;
						}
						inQueue.add(new Msg(Msg.INCOMING_MSG, msg, connection));
					}
				} catch (Exception e) {
					// notify dead connection
					inQueue.add(new Msg(Msg.CONNECTION_ENDED, null, connection));
				}
			}
		});
		recThread.start();

		// start the sending thread
		sendThread = new Thread(new Runnable() {

			@Override
			public void run() {
				Random rand = new Random();
				try {
					while (connection.keepRunning) {
						// wait for a msg to be placed in outQueue and then take it
						Msg msg = outQueue.poll(100, TimeUnit.MILLISECONDS);
						if (msg != null) {
							dataOut.writeInt(msg.msg.length);
							dataOut.write(msg.msg);
							dataOut.flush();
						}
					}
				} catch (Exception e) {
					// notify dead connection
					inQueue.add(new Msg(Msg.CONNECTION_ENDED, e.getMessage().getBytes(), connection));
				}
			}
		});
		sendThread.start();
		isActive = true;
	}

	/**
	 * Closes the connection's socket and stops the send and receive threads.
	 */
	public void close() {
		if (isActive) {
			try {
				keepRunning = false;
				try {
					socket.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
				recThread.interrupt();
				recThread.join();
				sendThread.interrupt();
				sendThread.join();
				isActive = false;
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
	}
}
