import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;

/**
 * 
 * This class provides support for handling AIO files.
 *
 */
public class FileOps {
	/*
	 * List all the files and folders inside the given directory.
	 */
	static ArrayList<String> listFiles(File dir, String prefix) {
		ArrayList<String> flist = new ArrayList<String>();
		if (!dir.exists() || !dir.isDirectory())
			return flist;
		prefix += dir.getName() + "/";
		flist.add(prefix);
		for (File file : dir.listFiles()) {
			if (file.isDirectory()) {
				flist.addAll(listFiles(file, prefix));
			} else {
				flist.add(prefix + file.getName());
			}
		}
		return flist;
	}

	/**
	 * Deletes the given directory.
	 * 
	 * @param dir : directory to delete
	 */
	public static void deleteDir(File dir) {
		if (!dir.exists() || !dir.isDirectory())
			return;
		for (File file : dir.listFiles()) {
			if (file.isDirectory())
				deleteDir(file);
			else
				file.delete();
		}
		dir.delete();
	}

	/**
	 * Create the whole folder with all the files inside from the given AIO file.
	 * 
	 * @param aio       : AIO file to create the folder from
	 * @param parentDir : create folder at this location
	 * @throws IOException
	 */
	public static void createFolderFromAIOFile(File aio, String parentDir) throws IOException {
		DataInputStream din = new DataInputStream(new FileInputStream(aio));
		int fileCount = din.readInt();
		for (int i = 0; i < fileCount; i++) {
			String fname = din.readUTF();
			int isDir = din.readInt();
			if (isDir == 1) {
				new File(parentDir + fname).mkdir();
			} else {

				DataOutputStream dout = new DataOutputStream(new FileOutputStream(new File(parentDir + fname)));
				long size = din.readLong();
				byte b[] = new byte[10240];
				while (size > 0) {
					int read = din.read(b, 0, (int) Math.min(size, b.length));
					dout.write(b, 0, read);
					size -= read;
				}
				dout.close();
			}
		}
		din.close();
	}

	/**
	 * Creates AIO file from the given folder.
	 * 
	 * @param srcDir : folder from which to create AIO file
	 * @param aio    : output AIO file
	 * @throws IOException
	 */
	public static void createAIOFile(File srcDir, File aio) throws IOException {
		ArrayList<String> fileList = listFiles(srcDir, "");
		DataOutputStream dout = new DataOutputStream(new FileOutputStream(aio));
		dout.writeInt(fileList.size());
		for (String s : fileList) {
			dout.writeUTF(s);
			String fullName = srcDir.getAbsoluteFile().getParent() + "/" + s;
			if (new File(fullName).isDirectory()) {
				dout.writeInt(1);
			} else {
				dout.writeInt(0);
				File srcFile = new File(fullName);
				long size = srcFile.length();
				dout.writeLong(size);
				DataInputStream din = new DataInputStream(new FileInputStream(srcFile));
				byte b[] = new byte[10240];
				while (size > 0) {
					int read = din.read(b);
					dout.write(b, 0, read);
					size -= read;
				}
				din.close();
			}
		}
		dout.close();
	}

}
