import java.io.BufferedReader;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;

/**
 * 
 * Represents the main thread server thread. It receives all the messages from
 * clients and forwards them to appropriate SyncThreads.
 *
 */

public class MainThread {
	int port;
	LinkedBlockingQueue<Msg> inbox;
	ListeningSocket listeningSocket;

	String dbFileName = "SyncServer.db";
	Database database;

	Map<String, SyncThread> syncThreadMap;

	static String mainFolderName = "FSync";
	static String tempFolderName = ".temp";
	String parentDir;

	String configFileName;

	/**
	 * 
	 * @param cfn : config file name
	 */
	public MainThread(String cfn) {
		configFileName = cfn;
		inbox = new LinkedBlockingQueue<Msg>();
		syncThreadMap = new HashMap<String, SyncThread>();
	}

	/**
	 * Reads the config file and sets port and working directory.
	 * 
	 * @throws IOException
	 */
	public void readConfig() throws IOException {
		BufferedReader in = new BufferedReader(new FileReader(new File(configFileName)));
		while (true) {
			String line = in.readLine();
			if (line == null)
				break;
			if (line.length() < 1 || line.startsWith("#"))
				continue;
			String tokens[] = line.split(":");
			if (tokens[0].equals("PORT")) {
				port = Integer.parseInt(tokens[1]);
			} else if (tokens[0].equals("OPERATION_DIR")) {
				parentDir = tokens[1];
			}
		}
		in.close();
	}

	/**
	 * Processes login request from client and replies accordingly.
	 * 
	 * @param msg : contains login request
	 * @return : username if valid login else null
	 * @throws IOException
	 */
	public String checkLogin(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		String uname = din.readUTF();
		String pass = din.readUTF();

		if (!database.checkLogin(uname, pass)) {
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.LOGIN_RESULT);
			dout.writeInt(Msg.LOGIN_ERROR);
			dout.writeUTF("Wrong username or password!");
			m.prepare();
			msg.connection.send(m);
			return null;
		} else {
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.LOGIN_RESULT);
			dout.writeInt(Msg.LOGIN_SUCCSESS);
			dout.writeUTF(uname);
			m.prepare();
			msg.connection.send(m);
			return uname;
		}
	}

	/**
	 * Processes signup request from client and replies accordingly.
	 * 
	 * @param msg : contains signup request
	 * @return : username if signup successful else null
	 * @throws IOException
	 */
	public String signup(Msg msg) throws IOException {
		DataInputStream din = msg.getDataInputStream();
		String uname = din.readUTF();
		String pass = din.readUTF();
		if (uname.length() < 1 || pass.length() < 1) {
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.SIGNUP_RESULT);
			dout.writeInt(Msg.SIGNUP_ERROR);
			dout.writeUTF("Username and password must not be empty!");
			m.prepare();
			msg.connection.send(m);
			return null;
		} else {
			if (!database.addUser(uname, pass)) {
				Msg m = new Msg(Msg.OUTGOING_MSG, null);
				DataOutputStream dout = m.getDataOutputStream();
				dout.writeInt(Msg.SIGNUP_RESULT);
				dout.writeInt(Msg.SIGNUP_ERROR);
				dout.writeUTF("Username taken!");
				m.prepare();
				msg.connection.send(m);
				return null;
			} else {
				Msg m = new Msg(Msg.OUTGOING_MSG, null);
				DataOutputStream dout = m.getDataOutputStream();
				dout.writeInt(Msg.SIGNUP_RESULT);
				dout.writeInt(Msg.SIGNUP_SUCCSESS);
				m.prepare();
				msg.connection.send(m);
				return uname;
			}
		}
	}

	/**
	 * Calls checkLogin and adds the client to appropriate sync thread if valid
	 * login. If sync thread for the client is not running start a new sync thread
	 * for the client.
	 * 
	 * @param msg : contains login request
	 * @throws IOException
	 */
	public void handleLogin(Msg msg) throws IOException {
		System.out.println("New login req");
		String uname = checkLogin(msg);
		if (uname != null) {
			System.out.println(uname + " logged in");
			msg.connection.userName = uname;
			SyncThread syncThread = syncThreadMap.get(uname);
			if (syncThread == null) {
				syncThread = new SyncThread(inbox, parentDir, uname);
				syncThreadMap.put(uname, syncThread);
				syncThread.keepRunning = true;
				syncThread.start();
			}
			syncThread.lock.lock();
			syncThread.clients.add(msg.connection);
			syncThread.lock.unlock();
		}
	}

	/**
	 * Calls signup and creates appropriate directories for the new user if signup
	 * is successful.
	 * 
	 * @param msg : contains signup request
	 * @throws IOException
	 */
	public void handleSignup(Msg msg) throws IOException {
		System.out.println("new signup req");
		String uname = signup(msg);
		if (uname != null) {
			System.out.println(uname + " signed up");
			File f;
			f = new File(parentDir + "userdata/" + uname + "/" + mainFolderName);
			f.mkdirs();
			f = new File(parentDir + "userdata/" + uname + "/" + tempFolderName);
			f.mkdirs();
		}
	}

	/**
	 * Reads config file, initializes database, starts listening socket and waits
	 * for the incoming messages.
	 * 
	 * @throws IOException
	 */
	public void start() throws IOException {
		readConfig();
		database = new Database(parentDir + dbFileName);
		listeningSocket = new ListeningSocket(port, inbox);
		listeningSocket.start();
		System.out.println("running on port : " + port);
		while (true) {
			Msg msg;
			try {
				msg = inbox.take();
				if (msg.type == Msg.CONNECTION_ENDED) {
					if (msg.connection.userName != null) {
						SyncThread syncThread = syncThreadMap.get(msg.connection.userName);
						syncThread.inbox.add(msg);
					}
				} else if (msg.type == Msg.SYNC_THREAD_ENDED) {
					SyncThread syncThread = syncThreadMap.get((String) msg.userData);
					syncThread.t.join();
					System.out.println("Joined sync thread, user : " + (String) msg.userData);
					syncThreadMap.remove((String) msg.userData);
				} else if (msg.type == Msg.INCOMING_MSG) {
					DataInputStream din = msg.getDataInputStream();
					int msgName = din.readInt();
					Connection conn = msg.connection;
					if (conn.userName == null) {
						if (msgName == Msg.LOGIN) {
							handleLogin(msg);
						} else if (msgName == Msg.SIGNUP) {
							handleSignup(msg);
						}
					} else {
						msg.resetDataInputStream();
						SyncThread syncThread = syncThreadMap.get(conn.userName);
						syncThread.inbox.add(msg);
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
	}

}