import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.security.DigestException;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;

/**
 * 
 * This objects represents the server side thread that handles all the sync
 * operations of all the machines belonging to the same user.
 *
 */
public class SyncThread {
	static int syncTimeout = 10 * 60 * 1000;

	LinkedBlockingQueue<Msg> inbox = new LinkedBlockingQueue<Msg>();
	LinkedBlockingQueue<Msg> mainThreadInbox;
	ArrayList<Connection> clients = new ArrayList<Connection>();
	ArrayList<Connection> syncRecipients = new ArrayList<Connection>();
	Connection syncSource = null;
	Lock lock = new ReentrantLock();
	boolean syncGoingOn = false;

	static String mainFolderName = MainThread.mainFolderName;
	static String tempFolderName = MainThread.tempFolderName;

	String parentDir;

	int fileNameCounter = 0;
	int syncCounter = 0;

	boolean keepRunning = false;
	Thread t = null;

	String userName;

	/**
	 * 
	 * @param mainThreadInbox : main thread inbox used to send messages to main
	 *                        thread
	 * @param parentDir       : working directory for the thread
	 * @param userName        : username of the the user this thread will handle
	 */
	public SyncThread(LinkedBlockingQueue<Msg> mainThreadInbox, String parentDir, String userName) {
		this.mainThreadInbox = mainThreadInbox;
		this.parentDir = parentDir;
		this.userName = userName;
	}

	void setAlarm(Msg msg, int milliSeconds) {
		new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					Thread.sleep(milliSeconds);
					inbox.add(msg);
				} catch (InterruptedException e) {
					e.printStackTrace();
				}
			}
		}).start();
	}

	/**
	 * Handles sync up request from client.
	 * 
	 * @param msg : message containing the request
	 * @throws IOException
	 * @throws DigestException
	 */
	public void handleSyncUpReq(Msg msg) throws IOException, DigestException {
		System.out.println("Got sync up req");
		if (syncGoingOn) {
			// deny the request
			System.out.println("sync up denied");
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.SYNC_UP_REQ_DENIED);
			dout.writeUTF("Another sync is in progress. Wait for some time and try again.");
			m.prepare();
			msg.connection.send(m);
		} else {
			// allow sync up
			syncCounter++;
			System.out.println("Sending sig file");
			syncGoingOn = true;
			syncSource = msg.connection;
			File aio = new File(parentDir + "userdata/" + msg.connection.userName + "/" + tempFolderName + "/aio");
			File srcFolder = new File(parentDir + "userdata/" + msg.connection.userName + "/" + mainFolderName);
			FileOps.createAIOFile(srcFolder, aio);
			Sync sync = new Sync();
			fileNameCounter++;
			msg.connection.sigFile = new File(parentDir + "userdata/" + msg.connection.userName + "/" + tempFolderName
					+ "/sigfile" + fileNameCounter);
			sync.generateSigFile(aio, msg.connection.sigFile);
			msg.connection.sendFile(msg.connection.sigFile, Msg.SIGFILE);
			msg.connection.sigFile.delete();
			msg.connection.sigFile = null;
			setAlarm(new Msg(Msg.END_SYNC_ALARM, Integer.valueOf(syncCounter)), syncTimeout);
		}
	}

	/**
	 * Handles sync down request from client.
	 * 
	 * @param msg : message containing the request
	 * @throws IOException
	 * @throws DigestException
	 */
	public void handleSyncDownReq(Msg msg) throws IOException {
		System.out.println("Got sync down req");
		if (syncGoingOn) {
			// deny request
			System.out.println("sync down denied");
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.SYNC_DOWN_REQ_DENIED);
			dout.writeUTF("Another sync is in progress. Wait for some time and try again.");
			m.prepare();
			msg.connection.send(m);
		} else {
			// allow sync down
			syncCounter++;
			System.out.println("sending sync down go ahead");
			syncGoingOn = true;
			syncRecipients = new ArrayList<Connection>();
			syncRecipients.add(msg.connection);
			Msg m = new Msg(Msg.OUTGOING_MSG, null);
			DataOutputStream dout = m.getDataOutputStream();
			dout.writeInt(Msg.SYNC_DOWN_REQ_GO_AHEAD);
			m.prepare();
			msg.connection.send(m);
			setAlarm(new Msg(Msg.END_SYNC_ALARM, Integer.valueOf(syncCounter)), syncTimeout);
		}
	}

	/**
	 * This is called when server receives a part of the signature file.
	 * 
	 * @param msg : contains a part of the signature file
	 * @throws IOException
	 * @throws InvalidSignatureFile
	 */
	public void processSigFile(Msg msg) throws IOException, InvalidSignatureFile {
		boolean flag = false;
		for (Connection c : syncRecipients) {
			if (c == msg.connection) {
				flag = true;
				break;
			}
		}
		if (!flag)
			return;
		String tempDir = parentDir + "userdata/" + msg.connection.userName + "/" + tempFolderName;
		if (msg.connection.sigFile == null) {
			// first part is received
			fileNameCounter++;
			msg.connection.sigFile = new File(tempDir + "/sigfile" + fileNameCounter);
			msg.connection.sigFileOut = new DataOutputStream(new FileOutputStream(msg.connection.sigFile));
		}
		DataInputStream din = msg.getDataInputStream();
		int isLastPart = din.readInt();
		int partSize = din.readInt();

		byte b[] = new byte[1024];
		while (partSize > 0) {
			int read = din.read(b);
			msg.connection.sigFileOut.write(b, 0, read);
			partSize -= read;
		}

		if (isLastPart == 1) {
			// sig file completely received
			System.out.println("sig file received");
			msg.connection.sigFileOut.close();
			File aio2 = new File(tempDir + "/aio2");
			if (!aio2.exists()) {
				FileOps.createAIOFile(
						new File(parentDir + "userdata/" + msg.connection.userName + "/" + mainFolderName), aio2);

			}
			Sync sync = new Sync();
			fileNameCounter++;
			msg.connection.deltaFile = new File(tempDir + "/deltafile" + fileNameCounter);
			sync.generateDeltaFile(aio2, msg.connection.sigFile, msg.connection.deltaFile);
			System.out.println("sending delta file");
			msg.connection.sendFile(msg.connection.deltaFile, Msg.DELTAFILE);
			msg.connection.sigFile.delete();
			msg.connection.sigFile = null;
			msg.connection.deltaFile.delete();
			msg.connection.deltaFile = null;
			syncRecipients.remove(msg.connection);
			if (syncRecipients.size() == 0) {
				aio2.delete();
				syncGoingOn = false;
				System.out.println("Sync done");
			}
		}
	}

	/**
	 * This is called when server receives a part of the delta file.
	 * 
	 * @param msg : contains a part of the delta file
	 * @throws IOException
	 * @throws InvalidSignatureFile
	 */
	public void processDeltaFile(Msg msg) throws IOException, InvalidSignatureFile {
		if (msg.connection != syncSource)
			return;
		String userDir = parentDir + "userdata/" + msg.connection.userName + "/";
		String tempDir = userDir + tempFolderName;
		String mainDir = userDir + mainFolderName;
		if (msg.connection.deltaFile == null) {
			// first part is received
			fileNameCounter++;
			msg.connection.deltaFile = new File(tempDir + "/deltafile" + fileNameCounter);
			msg.connection.deltaFileOut = new DataOutputStream(new FileOutputStream(msg.connection.deltaFile));
		}
		DataInputStream din = msg.getDataInputStream();
		int isLastPart = din.readInt();
		int partSize = din.readInt();

		byte b[] = new byte[1024];
		while (partSize > 0) {
			int read = din.read(b);
			msg.connection.deltaFileOut.write(b, 0, read);
			partSize -= read;
		}

		if (isLastPart == 1) {
			// delta file completely received
			System.out.println("delta file received");
			syncSource = null;
			msg.connection.deltaFileOut.close();
			File aio = new File(tempDir + "/aio");
			File aio2 = new File(tempDir + "/aio2");
			Sync sync = new Sync();
			sync.applyDelta(aio, msg.connection.deltaFile, aio2);
			FileOps.deleteDir(new File(mainDir));
			FileOps.createFolderFromAIOFile(aio2, userDir);
			aio.delete();
			msg.connection.deltaFile.delete();
			msg.connection.deltaFile = null;
			lock.lock();
			syncRecipients = new ArrayList<Connection>();
			for (Connection c : clients) {
				if (c != msg.connection) {
					syncRecipients.add(c);
				}
			}
			lock.unlock();
			if (syncRecipients.size() == 0) {
				System.out.println("no recipients, Sync done.");
				aio2.delete();
				syncGoingOn = false;
			} else {
				for (Connection c : syncRecipients) {
					Msg m = new Msg(Msg.OUTGOING_MSG, null);
					DataOutputStream dout = m.getDataOutputStream();
					dout.writeInt(Msg.INIT_SYNC_DOWN_INSTRUCTION_FROM_SERVER);
					m.prepare();
					c.send(m);
				}
			}
		}
	}

	/*
	 * cleans up closed connection
	 */
	void handleConnClosed(Msg msg) {
		lock.lock();
		clients.remove(msg.connection);
		int clientCount = clients.size();
		lock.unlock();
		if (syncRecipients.size() > 0) {
			syncRecipients.remove(msg.connection);
			if (syncRecipients.size() == 0) {
				String tempDir = parentDir + "userdata/" + msg.connection.userName + "/" + tempFolderName;
				File aio2 = new File(tempDir + "/aio2");
				aio2.delete();
				syncGoingOn = false;
				System.out.println("sync done");
			}
		}
		if (clientCount == 0) {
			Msg m = new Msg(Msg.SYNC_THREAD_ENDED, userName);
			mainThreadInbox.add(m);
			keepRunning = false;
		}
	}

	/**
	 * Starts a new thread that waits for the messages from client and processes
	 * them.
	 */
	public void start() {
		t = new Thread(new Runnable() {
			@Override
			public void run() {
				System.out.println("sync thread started, user : " + userName);
				while (keepRunning) {
					try {
						Msg msg = inbox.poll(100, TimeUnit.MILLISECONDS);
						if (msg == null)
							continue;
						if (msg.type == Msg.CONNECTION_ENDED) {
							handleConnClosed(msg);
						} else if (msg.type == Msg.END_SYNC_ALARM) {
							if (syncGoingOn && syncCounter == (Integer) msg.userData) {
								System.out.println("Sync timeout. Killing current sync.");
								syncGoingOn = false;
								syncSource = null;
								syncRecipients = new ArrayList<Connection>();
							}
						} else if (msg.type == Msg.INCOMING_MSG) {
							DataInputStream din = msg.getDataInputStream();
							int msgName = din.readInt();
							if (msgName == Msg.INIT_SYNC_UP_REQ_FROM_CLIENT) {
								handleSyncUpReq(msg);
							} else if (msgName == Msg.INIT_SYNC_DOWN_REQ_FROM_CLIENT) {
								handleSyncDownReq(msg);
							} else if (msgName == Msg.SIGFILE) {
								processSigFile(msg);
							} else if (msgName == Msg.DELTAFILE) {
								processDeltaFile(msg);
							}
						}
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
				System.out.println("sync thread done, user : " + userName);
			}
		});
		t.start();
	}
}